/*
	This is a file system recognizer for the Ext2 file system, release 3.
	Copyright (C) 1999	Bo Branten
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Please send comments, corrections and contributions to bosse@acc.umu.se

	The most recent version of this program is available from:
	http://www.acc.umu.se/~bosse/

	Revision history:

	3. Added some PAGED_CODE();

	2. Added ZwUnloadDriver on FSR.
       Reorganized CheckDeviceForExt2fs and ReadBlockDevice.

	1. Initial release.
*/

#include <ntifs.h>

#define DEVICE_NAME L"\\FileSystem\\Ext2fsRecognizer"

#define FSD_REGISTRY_PATH \
	L"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\Ext2fs"

#define FSR_REGISTRY_PATH \
	L"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\Ext2fs_Rec"

#define EXT2_SUPER_MAGIC		0xEF53
#define EXT2_SUPER_MAGIC_OFFSET	56
#define EXT2_SUPER_OFFSET		1024

NTSTATUS
DriverEntry (
	IN PDRIVER_OBJECT	DriverObject,
	IN PUNICODE_STRING	RegistryPath
);

VOID
DriverUnload (
	IN PDRIVER_OBJECT	DriverObject
);

NTSTATUS
FileSystemControl (
	IN PDEVICE_OBJECT	DeviceObject,
	IN PIRP				Irp
);

NTSTATUS
CheckDeviceForExt2fs (
	IN PDEVICE_OBJECT	DeviceObject
);

NTSTATUS
ReadBlockDevice (
	IN PDEVICE_OBJECT	DeviceObject,
	IN PLARGE_INTEGER	Offset,
	IN ULONG			Length,
	OUT PVOID			Buffer
);

NTSYSAPI
NTSTATUS
NTAPI
ZwLoadDriver (
	IN PUNICODE_STRING	RegistryPath
);

NTSYSAPI
NTSTATUS
NTAPI
ZwUnloadDriver (
	IN PUNICODE_STRING	RegistryPath
);

#pragma code_seg("init")

NTSTATUS
DriverEntry (
	IN PDRIVER_OBJECT	DriverObject,
	IN PUNICODE_STRING	RegistryPath
	)
{
	UNICODE_STRING	device_name;
	NTSTATUS		status;
	PDEVICE_OBJECT	device_object;

	KdPrint(("Ext2fs_Rec: DriverEntry\n"));

	DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] = FileSystemControl;

	RtlInitUnicodeString(&device_name, DEVICE_NAME);

	status = IoCreateDevice(
		DriverObject,
		0,
		&device_name,
		FILE_DEVICE_DISK_FILE_SYSTEM,
		0,
		FALSE,
		&device_object
		);

	if (NT_SUCCESS(status))
	{
		IoRegisterFileSystem(device_object);
	}

	MmPageEntireDriver(DriverEntry);

	return status;
}

#pragma code_seg("page")

VOID
DriverUnload (
	IN PDRIVER_OBJECT DriverObject
	)
{
	KdPrint(("Ext2fs_Rec: DriverUnload\n"));
}

NTSTATUS
FileSystemControl (
	IN PDEVICE_OBJECT	DeviceObject,
	IN PIRP				Irp
	)
{
	PIO_STACK_LOCATION	io_stack;
	NTSTATUS			status;
	UNICODE_STRING		registry_path;

	PAGED_CODE();

	io_stack = IoGetCurrentIrpStackLocation(Irp);

	ASSERT(io_stack != NULL);

	switch (io_stack->MinorFunction)
	{
	case IRP_MN_MOUNT_VOLUME:
		KdPrint(("Ext2fs_Rec: FileSystemControl: IRP_MN_MOUNT_VOLUME\n"));

		status = CheckDeviceForExt2fs(
			io_stack->Parameters.MountVolume.DeviceObject
			);

		Irp->IoStatus.Status = status;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);

		return status;

	case IRP_MN_LOAD_FILE_SYSTEM:
		KdPrint(("Ext2fs_Rec: FileSystemControl: IRP_MN_LOAD_FILE_SYSTEM\n"));

		RtlInitUnicodeString(&registry_path, FSD_REGISTRY_PATH);

		status = ZwLoadDriver(&registry_path);

		if (!NT_SUCCESS(status))
		{
			KdPrint(("Ext2fs_Rec: FileSystemControl: "
				"ZwLoadDriver failed with error 0x%x\n", status));
		}

		Irp->IoStatus.Status = status;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);

		IoUnregisterFileSystem(DeviceObject);

		DeviceObject->DriverObject->DriverUnload = DriverUnload;

		IoDeleteDevice(DeviceObject);

		RtlInitUnicodeString(&registry_path, FSR_REGISTRY_PATH);

		ZwUnloadDriver(&registry_path);

		return status;

	default:
		KdPrint(("Ext2fs_Rec: FileSystemControl: "
			"Unknown minor function code 0x%x\n", io_stack->MinorFunction));

		status = STATUS_NOT_IMPLEMENTED;

		Irp->IoStatus.Status = status;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest(Irp, IO_NO_INCREMENT);

		return status;
	}
}

NTSTATUS
CheckDeviceForExt2fs (
	IN PDEVICE_OBJECT DeviceObject
	)
{
	PUCHAR			buffer;
	LARGE_INTEGER	offset;
	NTSTATUS		status;
	USHORT			magic;

	PAGED_CODE();
	ASSERT(DeviceObject != NULL);

	buffer = (PUCHAR) ExAllocatePool(NonPagedPoolCacheAligned, 512);

	if (!buffer)
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	offset.QuadPart = EXT2_SUPER_OFFSET;

	status = ReadBlockDevice(
		DeviceObject,
		&offset,
		512,
		buffer
		);

	magic = *( (PUSHORT) &buffer[EXT2_SUPER_MAGIC_OFFSET] );

	ExFreePool(buffer);

	if (!NT_SUCCESS(status))
	{
		KdPrint(("Ext2fs_Rec: CheckDeviceForExt2fs: "
			"ReadBlockDevice failed with error 0x%x\n", status));
		return status;
	}
	else if (magic == EXT2_SUPER_MAGIC)
	{
		KdPrint(("Ext2fs_Rec: CheckDeviceForExt2fs: Ext2fs found\n"));
		return STATUS_FS_DRIVER_REQUIRED;
	}
	else
	{
		KdPrint(("Ext2fs_Rec: CheckDeviceForExt2fs: Ext2fs not found\n"));
		return STATUS_UNRECOGNIZED_VOLUME;
	}
}

NTSTATUS
ReadBlockDevice (
	IN PDEVICE_OBJECT	DeviceObject,
	IN PLARGE_INTEGER	Offset,
	IN ULONG			Length,
	OUT PVOID			Buffer
	)
{
	KEVENT			event;
	PIRP			irp;
	IO_STATUS_BLOCK io_status;
	NTSTATUS		status;

	PAGED_CODE();
	ASSERT(DeviceObject != NULL);
	ASSERT(Offset != NULL);
	ASSERT(Buffer != NULL);

	KeInitializeEvent(&event, NotificationEvent, FALSE);

	irp = IoBuildSynchronousFsdRequest(
		IRP_MJ_READ,
		DeviceObject,
		Buffer,
		Length,
		Offset,
		&event,
		&io_status
		);

	if (!irp)
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	status = IoCallDriver(DeviceObject, irp);

	if (status == STATUS_PENDING)
	{
		KeWaitForSingleObject(
			&event,
			Suspended,
			KernelMode,
			FALSE,
			NULL
			);
		status = io_status.Status;
	}

	return status;
}
